@@ -106,6 +106,7 @@ typedef struct {
         gpointer completion_data;
         SoupHTTP2IOState state;
         GError *error;
+        uint32_t http2_error;
         gboolean paused;
         guint32 stream_id;
         gboolean can_be_restarted;
@@ -170,6 +171,20 @@ set_error_for_data (SoupHTTP2MessageData *data,
                 g_error_free (error);
 }
 
+static void
+set_http2_error_for_data (SoupHTTP2MessageData *data,
+                          uint32_t              error_code)
+{
+        h2_debug (data->io, data, "[SESSION] Error: %s", nghttp2_http2_strerror (error_code));
+
+        if (data->error)
+                return;
+
+        data->http2_error = error_code;
+        data->error = g_error_new (G_IO_ERROR, G_IO_ERROR_FAILED,
+                                   "HTTP/2 Error: %s", nghttp2_http2_strerror (error_code));
+}
+
 static void
 set_io_error (SoupClientMessageIOHTTP2 *io,
               GError                   *error)
@@ -180,6 +195,11 @@ set_io_error (SoupClientMessageIOHTTP2 *io,
                 io->error = error;
         else
                 g_error_free (error);
+
+        if (io->close_task && !io->goaway_sent) {
+                g_task_return_boolean (io->close_task, TRUE);
+                g_clear_object (&io->close_task);
+        }
 }
 
 static void
@@ -206,6 +226,22 @@ advance_state_from (SoupHTTP2MessageData *data,
         data->state = to;
 }
 
+static gboolean
+soup_http2_message_data_can_be_restarted (SoupHTTP2MessageData *data,
+                                          GError               *error)
+{
+        if (data->can_be_restarted)
+                return TRUE;
+
+        return data->state < STATE_READ_DATA_START &&
+                !g_error_matches (error, G_IO_ERROR, G_IO_ERROR_TIMED_OUT) &&
+                !g_error_matches (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK) &&
+                !g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED) &&
+                error->domain != G_TLS_ERROR &&
+                data->http2_error == NGHTTP2_NO_ERROR &&
+                SOUP_METHOD_IS_IDEMPOTENT (soup_message_get_method (data->msg));
+}
+
 static void
 soup_http2_message_data_check_status (SoupHTTP2MessageData *data)
 {
@@ -217,6 +253,7 @@ soup_http2_message_data_check_status (SoupHTTP2MessageData *data)
         if (g_cancellable_set_error_if_cancelled (g_task_get_cancellable (task), &error)) {
                 io->pending_io_messages = g_list_remove (io->pending_io_messages, data);
                 data->task = NULL;
+                soup_client_message_io_http2_finished ((SoupClientMessageIO *)io, msg);
                 g_task_return_error (task, error);
                 g_object_unref (task);
                 return;
@@ -231,7 +268,7 @@ soup_http2_message_data_check_status (SoupHTTP2MessageData *data)
         if (data->error) {
                 GError *error = g_steal_pointer (&data->error);
 
-                if (data->can_be_restarted)
+                if (soup_http2_message_data_can_be_restarted (data, error))
                         data->item->state = SOUP_MESSAGE_RESTARTING;
                 else
                         soup_message_set_metrics_timestamp (data->msg, SOUP_MESSAGE_METRICS_RESPONSE_END);
@@ -300,7 +337,7 @@ io_write_ready (GObject                  *stream,
                 return G_SOURCE_REMOVE;
         }
 
-        while (nghttp2_session_want_write (io->session) && !error)
+        while (!error && nghttp2_session_want_write (io->session))
                 io_write (io, FALSE, NULL, &error);
 
         if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK)) {
@@ -328,7 +365,7 @@ io_try_write (SoupClientMessageIOHTTP2 *io,
                 if (blocking || !nghttp2_session_want_write (io->session))
                         return;
         } else {
-                while (nghttp2_session_want_write (io->session) && !error)
+                while (!error && nghttp2_session_want_write (io->session))
                         io_write (io, blocking, NULL, &error);
         }
 
@@ -336,6 +373,8 @@ io_try_write (SoupClientMessageIOHTTP2 *io,
                 g_clear_error (&error);
                 io->write_source = g_pollable_output_stream_create_source (G_POLLABLE_OUTPUT_STREAM (io->ostream), NULL);
                 g_source_set_name (io->write_source, "Soup HTTP/2 write source");
+                /* Give write more priority than read */
+                g_source_set_priority (io->write_source, G_PRIORITY_DEFAULT - 1);
                 g_source_set_callback (io->write_source, (GSourceFunc)io_write_ready, io, NULL);
                 g_source_attach (io->write_source, g_main_context_get_thread_default ());
         }
@@ -354,14 +393,24 @@ io_read (SoupClientMessageIOHTTP2  *io,
         gssize read;
         int ret;
 
+        /* Always try to write before read, in case there's a pending reset stream after an error. */
+        io_try_write (io, blocking);
+
         if ((read = g_pollable_stream_read (io->istream, buffer, sizeof (buffer),
                                             blocking, cancellable, error)) < 0)
             return FALSE;
 
+        if (read == 0) {
+                g_set_error_literal (error, G_IO_ERROR,
+                                     G_IO_ERROR_PARTIAL_INPUT,
+                                     _("Connection terminated unexpectedly"));
+                return FALSE;
+        }
+
         g_warn_if_fail (io->in_callback == 0);
         ret = nghttp2_session_mem_recv (io->session, buffer, read);
         NGCHECK (ret);
-        return ret != 0;
+        return ret > 0;
 }
 
 static gboolean
@@ -384,7 +433,7 @@ io_read_ready (GObject                  *stream,
         if (conn)
                 soup_connection_set_in_use (conn, TRUE);
 
-        while (nghttp2_session_want_read (io->session) && progress) {
+        while (progress && nghttp2_session_want_read (io->session)) {
                 progress = io_read (io, FALSE, NULL, &error);
                 if (progress) {
                         g_list_foreach (io->pending_io_messages,
@@ -402,6 +451,8 @@ io_read_ready (GObject                  *stream,
                 return G_SOURCE_CONTINUE;
         }
 
+        io->is_shutdown = TRUE;
+
         if (error) {
                 set_io_error (io, error);
                 g_list_foreach (io->pending_io_messages,
@@ -409,8 +460,6 @@ io_read_ready (GObject                  *stream,
                                 NULL);
         }
 
-        io->is_shutdown = TRUE;
-
         g_clear_pointer (&io->read_source, g_source_unref);
         if (conn) {
                 soup_connection_set_in_use (conn, FALSE);
@@ -572,8 +621,7 @@ handle_goaway (SoupClientMessageIOHTTP2 *io,
                 if ((error_code == 0 && (int32_t)data->stream_id > last_stream_id) ||
                      data->state < STATE_READ_DONE) {
                         /* TODO: We can restart unfinished messages */
-                        set_error_for_data (data, g_error_new (G_IO_ERROR, G_IO_ERROR_FAILED,
-                                            "HTTP/2 Error: %s", nghttp2_http2_strerror (error_code)));
+                        set_http2_error_for_data (data, error_code);
                 }
         }
 }
@@ -687,10 +735,8 @@ on_frame_recv_callback (nghttp2_session     *session,
                 }
                 break;
         case NGHTTP2_RST_STREAM:
-                if (frame->rst_stream.error_code != NGHTTP2_NO_ERROR) {
-                        set_error_for_data (data, g_error_new_literal (G_IO_ERROR, G_IO_ERROR_FAILED,
-                                                                       nghttp2_http2_strerror (frame->rst_stream.error_code)));
-                }
+                if (frame->rst_stream.error_code != NGHTTP2_NO_ERROR)
+                        set_http2_error_for_data (data, frame->rst_stream.error_code);
                 break;
         case NGHTTP2_WINDOW_UPDATE:
                 h2_debug (io, data, "[RECV] WINDOW_UPDATE: increment=%d, total=%d", frame->window_update.window_size_increment,
@@ -910,14 +956,19 @@ on_stream_close_callback (nghttp2_session *session,
         data->io->in_callback++;
 
         switch (error_code) {
+        case NGHTTP2_NO_ERROR:
+                break;
         case NGHTTP2_REFUSED_STREAM:
-                if (data->state < STATE_READ_DATA)
+                if (data->state < STATE_READ_DATA_START)
                         data->can_be_restarted = TRUE;
                 break;
         case NGHTTP2_HTTP_1_1_REQUIRED:
                 soup_message_set_force_http_version (data->item->msg, SOUP_HTTP_1_1);
                 data->can_be_restarted = TRUE;
                 break;
+        default:
+                set_http2_error_for_data (data, error_code);
+                break;
         }
 
         data->io->in_callback--;
@@ -1335,6 +1386,7 @@ soup_client_message_io_http2_finished (SoupClientMessageIO *iface,
 	SoupMessageIOCompletionFn completion_cb;
 	gpointer completion_data;
         SoupMessageIOCompletion completion;
+        gboolean is_closed;
 
         data = get_data_for_message (io, msg);
 
@@ -1347,9 +1399,10 @@ soup_client_message_io_http2_finished (SoupClientMessageIO *iface,
 
 	g_object_ref (msg);
 
+        is_closed = nghttp2_session_get_stream_user_data (io->session, data->stream_id) == NULL;
         nghttp2_session_set_stream_user_data (io->session, data->stream_id, NULL);
 
-        if (!io->is_shutdown) {
+        if (!io->is_shutdown && !is_closed) {
                 SoupConnection *conn;
 
                 NGCHECK (nghttp2_submit_rst_stream (io->session, NGHTTP2_FLAG_NONE, data->stream_id,
@@ -1551,11 +1604,13 @@ io_run_until (SoupClientMessageIOHTTP2 *io,
                 progress = io_run (data, cancellable, &my_error);
 
         if (my_error) {
-                g_propagate_error (error, my_error);
-                g_object_unref (msg);
-                return FALSE;
+                io->is_shutdown = TRUE;
+                set_io_error (io, my_error);
         }
 
+        if (io->error && !data->error)
+                data->error = g_error_copy (io->error);
+
 	if (data->error) {
                 g_propagate_error (error, g_steal_pointer (&data->error));
 		g_object_unref (msg);
@@ -1584,12 +1639,13 @@ soup_client_message_io_http2_run_until_read (SoupClientMessageIO  *iface,
 {
         SoupClientMessageIOHTTP2 *io = (SoupClientMessageIOHTTP2 *)iface;
         SoupHTTP2MessageData *data = get_data_for_message (io, msg);
+        GError *my_error = NULL;
 
-        if (io_run_until (io, msg, STATE_READ_DATA, cancellable, error))
+        if (io_run_until (io, msg, STATE_READ_DATA, cancellable, &my_error))
                 return TRUE;
 
         if (get_io_data (msg) == io) {
-                if (data->can_be_restarted)
+                if (soup_http2_message_data_can_be_restarted (data, my_error))
                         data->item->state = SOUP_MESSAGE_RESTARTING;
                 else
                         soup_message_set_metrics_timestamp (msg, SOUP_MESSAGE_METRICS_RESPONSE_END);
@@ -1597,6 +1653,8 @@ soup_client_message_io_http2_run_until_read (SoupClientMessageIO  *iface,
                 soup_client_message_io_http2_finished (iface, msg);
         }
 
+        g_propagate_error (error, my_error);
+
         return FALSE;
 }
 
@@ -1670,6 +1728,7 @@ soup_client_message_io_http2_set_owner (SoupClientMessageIOHTTP2 *io,
 
         io->read_source = g_pollable_input_stream_create_source (G_POLLABLE_INPUT_STREAM (io->istream), NULL);
         g_source_set_name (io->read_source, "Soup HTTP/2 read source");
+        g_source_set_priority (io->read_source, G_PRIORITY_DEFAULT);
         g_source_set_callback (io->read_source, (GSourceFunc)io_read_ready, io, NULL);
         g_source_attach (io->read_source, g_main_context_get_thread_default ());
 }
@@ -1692,7 +1751,7 @@ soup_client_message_io_http2_close_async (SoupClientMessageIO *iface,
 
         soup_client_message_io_http2_terminate_session (io);
         if (!io->async) {
-                g_assert (io->goaway_sent);
+                g_assert (io->goaway_sent || io->error);
                 return FALSE;
         }
 
